@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2000-2003 Marc Alexander Lehmann <pcg@goof.com>
+ * Copyright (c) 2000-2005 Marc Alexander Lehmann <schmorp@schmorp.de>
  * 
  * Redistribution and use in source and binary forms, with or without modifica-
  * tion, are permitted provided that the following conditions are met:
@@ -45,6 +45,13 @@
 # define SET_ERRNO(n) errno = (n)
 #endif
 
+#if (__i386 || __amd64) && __GNUC__ >= 3
+# define lzf_movsb(dst, src, len)                \
+   asm ("rep movsb"                              \
+        : "=D" (dst), "=S" (src), "=c" (len)     \
+        :  "0" (dst),  "1" (src),  "2" (len));
+#endif
+
 unsigned int 
 lzf_decompress (const void *const in_data,  unsigned int in_len,
                 void             *out_data, unsigned int out_len)
@@ -68,10 +75,16 @@ lzf_decompress (const void *const in_data,  unsigned int in_len,
               return 0;
             }
 
-#if USE_MEMCPY
-          memcpy (op, ip, ctrl);
-          op += ctrl;
-          ip += ctrl;
+#if CHECK_INPUT
+          if (ip + ctrl > in_end)
+            {
+              SET_ERRNO (EINVAL);
+              return 0;
+            }
+#endif
+
+#ifdef lzf_movsb
+          lzf_movsb (op, ip, ctrl);
 #else
           do
             *op++ = *ip++;
@@ -84,9 +97,25 @@ lzf_decompress (const void *const in_data,  unsigned int in_len,
 
           u8 *ref = op - ((ctrl & 0x1f) << 8) - 1;
 
+#if CHECK_INPUT
+          if (ip >= in_end)
+            {
+              SET_ERRNO (EINVAL);
+              return 0;
+            }
+#endif
           if (len == 7)
-            len += *ip++;
-          
+            {
+              len += *ip++;
+#if CHECK_INPUT
+              if (ip >= in_end)
+                {
+                  SET_ERRNO (EINVAL);
+                  return 0;
+                }
+#endif
+            }
+
           ref -= *ip++;
 
           if (op + len + 2 > out_end)
@@ -101,15 +130,20 @@ lzf_decompress (const void *const in_data,  unsigned int in_len,
               return 0;
             }
 
+#ifdef lzf_movsb
+          len += 2;
+          lzf_movsb (op, ref, len);
+#else
           *op++ = *ref++;
           *op++ = *ref++;
 
           do
             *op++ = *ref++;
           while (--len);
+#endif
         }
     }
-  while (op < out_end && ip < in_end);
+  while (ip < in_end);
 
   return op - (u8 *)out_data;
 }
